local MODNAME = 'SymphoniousCombat'
local ShivvAudio = LibStub('AceAddon-3.0'):GetAddon('ShivvAudio')
local SymphoniousCombat = ShivvAudio:NewModule(MODNAME, 'AceEvent-3.0')

local SCData = ShivvAudio:GetModule('SymphoniousCombatData')
local currentZone, zoneBosses

local sc
local knownMobs, activeMobs, inTier = {}, 0, 1
local combatStartFunc, combatExitFunc

local checkedLogEvents = {
	['RANGE_DAMAGE'] = true,
	['RANGE_MISSED'] = true,
	['SPELL_DAMAGE'] = true,
	['SPELL_MISSED'] = true,
	['SPELL_CAST_SUCCESS'] = true,
	['SPELL_PERIODIC_DAMAGE'] = true,
	['SWING_DAMAGE'] = true,
	['SWING_MISSED'] = true,
}

local combatMusic = {
	['shivvInternalCombatOutside1'] = true,
	['shivvInternalCombatOutside2'] = true,
	['shivvInternalCombatOutside3'] = true,
	['shivvInternalCombatOutsideElite'] = true,
	['shivvInternalCombatParty1'] = true,
	['shivvInternalCombatParty2'] = true,
	['shivvInternalCombatParty3'] = true,
	['shivvInternalCombatPartyBoss'] = true,
	['shivvInternalCombatBattleground'] = true,
	['shivvInternalCombatRaidTrash'] = true,
	['shivvInternalCombatRaidBoss'] = true
}

function SymphoniousCombat:OnInitialize()
	sc = ShivvAudio.db.profile.symphoniousCombat -- local db ref
	self:SetEnabledState(ShivvAudio.db.profile.enableSymphoniousCombat)
end

function SymphoniousCombat:OnEnable()
	ShivvAudio:EnableModule('SymphoniousCombatData') -- load up the data module

	-- if enabled, we'll need the base events
	self:RegisterEvent('PLAYER_ENTERING_WORLD')
	self:RegisterEvent('PLAYER_REGEN_DISABLED')
	self:RegisterEvent('PLAYER_REGEN_ENABLED')

	if (sc.advanced) then -- if using advanced mode then activate additional tracking events
		-- clear data tables
		for k in pairs(knownMobs) do knownMobs[k] = nil end

		self:PLAYER_ENTERING_WORLD() -- call to update nessecary location-based functions
	end
end

function SymphoniousCombat:OnDisable()
	ShivvAudio:DisableModule('SymphoniousCombatData') -- disable the data module

	self:UnregisterEvent('PLAYER_REGEN_ENABLED')
	self:UnregisterEvent('PLAYER_REGEN_DISABLED')
	self:UnregisterEvent('PLAYER_ENTERING_WORLD')
	self:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED') -- just to make sure its gone as its expensive
end

-- ----------------------- --
-- COMBAT START/EXIT FUNCS --
-- ----------------------- --
function SymphoniousCombat:EnterCombatBattleground()
	ShivvAudio:SoundHandler(sc.audio.battlegroundenter)
	ShivvAudio:MusicHandler('shivvInternalCombatBattleground', true)
end

function SymphoniousCombat:ExitCombatBattleground()
	ShivvAudio:SoundHandler(sc.audio.battlegroundleave)
	ShivvAudio:MusicHandler('shivvInternalCombatBattleground', false)
end

---------------------------------------------------------------

function SymphoniousCombat:EnterCombatParty()
	ShivvAudio:SoundHandler(sc.audio.partyenter)
	ShivvAudio:MusicHandler('shivvInternalCombatParty1', true)

	if (sc.advanced) then
		activeMobs = 0
		inTier = 1

		for g in pairs(knownMobs) do knownMobs[g] = nil end

		self:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED', 'LogEventParty')
	end
end

do
	local IS_FRIEND = COMBATLOG_OBJECT_AFFILIATION_MINE + COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID
	local IS_HOSTILENPC = bit.bor(COMBATLOG_OBJECT_TYPE_NPC, COMBATLOG_OBJECT_CONTROL_NPC, COMBATLOG_OBJECT_REACTION_HOSTILE) --0x00000240
	local IS_NULLGUID = 0x0000000000000000
	local band = bit.band
	local hostileMob

	function SymphoniousCombat:LogEventParty(cleu, time, evt, hc, sg, sn, sf, sf2, dg, dn, df)
		if (not checkedLogEvents[evt] or not dg) then return end

		hostileMob = nil

		if (band(sf, IS_FRIEND) ~= 0 and band(df, IS_HOSTILENPC) ~= 0) then
			hostileMob = dg
		elseif (band(sf, IS_HOSTILENPC) ~= 0 and band(df, IS_FRIEND) ~= 0) then
			hostileMob = sg
		end

		if (knownMobs[hostileMob] or hostileMob == nil or hostileMob == IS_NULLGUID) then return end

		if (zoneBosses[tonumber(strsub(hostileMob, 7, 10), 16)]) then
			ShivvAudio:MusicHandler('shivvInternalCombatPartyBoss', true)
			self:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
		else
			knownMobs[hostileMob] = true
			activeMobs = activeMobs + 1

			if (activeMobs > sc.partyMaxT2 and inTier < 3) then -- enough for t3 and not already playing it
				inTier = 3
				-- play t3 music
				ShivvAudio:MusicHandler('shivvInternalCombatParty3', true)
			elseif (activeMobs > sc.partyMaxT1 and inTier < 2) then -- enough for t2 and not already playing it (or higher)
				inTier = 2
				-- play t2 music
				ShivvAudio:MusicHandler('shivvInternalCombatParty2', true)
			end
		end
	end
end

function SymphoniousCombat:ExitCombatParty()
	ShivvAudio:SoundHandler(sc.audio.partyleave)

	if (sc.advanced) then
		self:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end

	ShivvAudio:MusicMultiKill(combatMusic)
end

---------------------------------------------------------------

function SymphoniousCombat:EnterCombatRaid()
	ShivvAudio:SoundHandler(sc.audio.raidenter)
	ShivvAudio:MusicHandler('shivvInternalCombatRaidTrash', true)

	if (sc.advanced) then
		for g in pairs(knownMobs) do knownMobs[g] = nil end

		self:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED', 'LogEventRaid')
	end
end

do
	local IS_FRIEND = COMBATLOG_OBJECT_AFFILIATION_MINE + COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID
	local IS_HOSTILENPC = bit.bor(COMBATLOG_OBJECT_TYPE_NPC, COMBATLOG_OBJECT_CONTROL_NPC, COMBATLOG_OBJECT_REACTION_HOSTILE) --0x00000240
	local IS_NULLGUID = 0x0000000000000000
	local band = bit.band
	local hostileMob

	function SymphoniousCombat:LogEventRaid(cleu, time, evt, hc, sg, sn, sf, sf2, dg, dn, df)
		if (not checkedLogEvents[evt] or not dg) then return end

		hostileMob = nil

		if (band(sf, IS_FRIEND) > 0 and band(df, IS_HOSTILENPC) > 0) then
			hostileMob = dg
		elseif (band(sf, IS_HOSTILENPC) > 0 and band(df, IS_FRIEND) > 0) then
			hostileMob = sg
		end

		if (knownMobs[hostileMob] or hostileMob == nil or hostileMob == IS_NULLGUID) then return end

		if (zoneBosses[tonumber(strsub(hostileMob, 7, 10), 16)]) then
			ShivvAudio:MusicHandler('shivvInternalCombatRaidBoss', true)
			self:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
		else
			knownMobs[hostileMob] = true
		end
	end
end

function SymphoniousCombat:ExitCombatRaid()
	ShivvAudio:SoundHandler(sc.audio.raidleave)

	if (sc.advanced) then
		self:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end

	ShivvAudio:MusicMultiKill(combatMusic)
end

---------------------------------------------------------------

local eliteMobs = {}

function SymphoniousCombat:EnterCombatOutside()
	ShivvAudio:SoundHandler(sc.audio.outsideenter)

	if (sc.advanced) then
		self:ScanForElites() -- populate elite list with any targeted elites

		activeMobs = 0
		inTier = 1

		for g in pairs(knownMobs) do knownMobs[g] = nil end

		ShivvAudio:MusicHandler('shivvInternalCombatOutside1', true)
		self:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED', 'LogEventOutside')
	else
		ShivvAudio:MusicHandler('shivvInternalCombatOutside1', true)
	end
end

do
	local IS_FRIEND = COMBATLOG_OBJECT_AFFILIATION_MINE + COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID
	local IS_HOSTILENPC = bit.bor(COMBATLOG_OBJECT_TYPE_NPC, COMBATLOG_OBJECT_CONTROL_NPC, COMBATLOG_OBJECT_REACTION_HOSTILE) --0x00000240
	local IS_NULLGUID = 0x0000000000000000
	local band = bit.band
	local hostileMob

	function SymphoniousCombat:LogEventOutside(cleu, time, evt, hc, sg, sn, sf, sf2, dg, dn, df)
		if (not checkedLogEvents[evt] or not dn) then return end

		hostileMob = nil

		if (band(sf, IS_FRIEND) ~= 0 and band(df, IS_HOSTILENPC) ~= 0) then
			hostileMob = dg
		elseif (band(sf, IS_HOSTILENPC) ~= 0 and band(df, IS_FRIEND) ~= 0) then
			hostileMob = sg
		end

		if (knownMobs[hostileMob] or hostileMob == nil or hostileMob == IS_NULLGUID) then return end

		self:ScanForElites()

		if (eliteMobs[tonumber(strsub(hostileMob, 7, 10), 16)]) then
			ShivvAudio:MusicHandler('shivvInternalCombatOutsideElite', true)
			self:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
		else
			knownMobs[hostileMob] = true
			activeMobs = activeMobs + 1

			if (activeMobs > sc.outsideMaxT2 and inTier < 3) then -- enough for t3 and not already playing it
				inTier = 3
				-- play t3 music
				ShivvAudio:MusicHandler('shivvInternalCombatOutside3', true)
			elseif (activeMobs > sc.outsideMaxT1 and inTier < 2) then -- enough for t2 and not already playing it (or higher)
				inTier = 2
				-- play t2 music
				ShivvAudio:MusicHandler('shivvInternalCombatOutside2', true)
			end
		end
	end
end

function SymphoniousCombat:PLAYER_TARGET_CHANGED()
	self:ScanTargetForElites()
end

local function sizeOf(tbl)
	local size = 0
	for k in pairs(tbl) do
		size = size + 1
	end
	return size
end

do
	local validElites = {
		['worldboss'] = true,
		['rareelite'] = true,
		['elite'] = true,
		['rare'] = true
	}
	local unitsToScan = {
		['target'] = true,
		['pettarget'] = true,
		['party1target'] = true,
		['party2target'] = true,
		['party3target'] = true,
		['party4target'] = true,
	}
	local UnitExists, UnitClassification, UnitGUID = _G.UnitExists, _G.UnitClassification, _G.UnitGUID

	function SymphoniousCombat:ScanForElites()
		for unit in pairs(unitsToScan) do
			if (UnitExists(unit) and validElites[UnitClassification(unit)]) then
				eliteMobs[tonumber(strsub(UnitGUID(unit), 7, 10), 16)] = true
			end
		end
	end

	function SymphoniousCombat:ScanTargetForElites()
		if (UnitExists('target') and validElites[UnitClassification('target')]) then
			eliteMobs[tonumber(strsub(UnitGUID('target'), 7, 10), 16)] = true
		end
	end
end

function SymphoniousCombat:ExitCombatOutside()
	ShivvAudio:SoundHandler(sc.audio.outsideleave)

	if (sc.advanced) then
		self:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end

	ShivvAudio:MusicMultiKill(combatMusic)
end

-- -------------- --
-- EVENT HANDLERS --
-- -------------- --
function SymphoniousCombat:PLAYER_ENTERING_WORLD()
	inInstance, instanceType = IsInInstance()

	self:UnregisterEvent('PLAYER_TARGET_CHANGED') -- kill target event (only used outside)

	if (inInstance) then
		if (instanceType == 'pvp' and sc.enableBattleground) then
			combatStartFunc, combatExitFunc = 'EnterCombatBattleground', 'ExitCombatBattleground'
			return
		elseif (instanceType == 'party' and sc.enableParty) then
			combatStartFunc, combatExitFunc = 'EnterCombatParty', 'ExitCombatParty'
			return
		elseif (instanceType == 'raid' and sc.enableRaid) then
			combatStartFunc, combatExitFunc = 'EnterCombatRaid', 'ExitCombatRaid'
			return
		end
	else
		if (sc.enableOutside) then
			combatStartFunc, combatExitFunc = 'EnterCombatOutside' , 'ExitCombatOutside'

			self:RegisterEvent('PLAYER_TARGET_CHANGED')

			return
		end
	end

	-- in arena or not enabled for current area type
	combatStartFunc, combatExitFunc = nil, nil
end

function SymphoniousCombat:PLAYER_REGEN_DISABLED()
	if (GetCurrentMapAreaID() ~= currentZone) then -- different zone from cache, update
		currentZone = GetCurrentMapAreaID()
		zoneBosses = SCData:GetZoneData(currentZone)
	end

	if (combatStartFunc and combatExitFunc) then
		self[combatStartFunc](self)
	end
end

function SymphoniousCombat:PLAYER_REGEN_ENABLED()
	if (combatStartFunc and combatExitFunc) then
		self[combatExitFunc](self)
	end
end
